from pathlib import Path

import numpy as np
import tifffile
import zarr
from tqdm import tqdm

from labnas.local.base import list_tif_files, sort_tif_files


def convert_to_tif(source_folder: Path, target_file: Path) -> None:
    """Convert a folder with multiple tifs into a single tif file."""
    if not source_folder.is_dir():
        raise FileNotFoundError(f"{source_folder} is not a directory.")
    if target_file.is_dir():
        raise FileExistsError(f"{target_file} already exists.")
    tif_files = list_tif_files(source_folder)
    tif_files = sort_tif_files(tif_files)
    n_files = tif_files.size
    for i_file in tqdm(range(n_files)):
        image = tifffile.imread(tif_files[i_file])
        if i_file == 0:
            shape = (n_files, image.shape[0], image.shape[1])
            tifffile.imwrite(
                target_file,
                shape=shape,
                dtype=np.uint8,
            )
            store = tifffile.imread(target_file, mode="r+", aszarr=True)
            z = zarr.open(store, mode="r+")
            print(f"Empty tif created: {shape}")
        z[i_file, :, :] = image
    store.close()


"""NAS interactions specific to imaging data."""
import os
import re
from pathlib import Path

import numpy as np
import tifffile
import zarr
from tqdm import tqdm
from tifffile.tifffile import TiffFileError

from labnas.remote.base import SftpNas
from labnas.local.base import identify_tif_origin, TifFileMissingException
from labnas.local.base import assert_single_stack, sort_tif_files


class ImagingNas(SftpNas):
    """Extends SftpNas for imaging data."""

    def sort_folder(
            self,
            remote_folder: Path,
            local_temp_folder: Path,
            remote_trash_folder: Path,
            recursive: bool = True,
            level: int = 0,
    ) -> None:
        """Sort a folder with imaging .tif files."""
        self.logger.info(f"=> Investigating: {remote_folder} (Level: {level})")
        if "INCOMPLETE" in remote_folder.name:
            self.logger.info(f"Skipping {remote_folder} -> INCOMPLETE flag.")
        elif "multipage" in remote_folder.name:
            self.logger.info(f"Skipping {remote_folder} -> multipage flag.")
        else:
            stacks, single_tifs = self.check_folder_structure(remote_folder)
            for stack_name, stack_tifs in stacks.items():
                self.process_stack(
                    stack_name=stack_name,
                    stack_tifs=stack_tifs,
                    remote_folder=remote_folder,
                    local_temp_folder=local_temp_folder,
                    remote_trash_folder=remote_trash_folder,
                )
            if recursive:
                _, folders = self.list_files_and_folders(remote_folder)
                for folder in folders:
                    self.sort_folder(folder, local_temp_folder, remote_trash_folder, level=level + 1)

    def process_stack(self, stack_name: str, stack_tifs: list, remote_folder: Path, local_temp_folder: Path, remote_trash_folder: Path) -> None:
        """
        1. Download stack as multipage tif.
        2. Upload multipage tif.
        3. Delete local multipage tif.
        4. Delete nas stack tifs.
        """
        # set paths
        remote_tif = remote_folder / "multipage_tiff" / f"{stack_name}.tif"
        self.logger.info(f"---{remote_tif}---")
        if self.is_file(remote_tif):
            raise FileExistsError(f"{remote_tif}")
        if not self.is_dir(remote_tif.parent):
            self.create_empty_folder(remote_tif.parent)
            self.logger.info(f"{remote_tif.parent} created.")
        self.logger.info(f"Downloading {stack_name} as as single tif.")

        trash_name = remote_folder / stack_name
        trash_name = str(trash_name).replace("/", "_")
        stack_trash_folder = remote_trash_folder / trash_name
        if self.is_file(stack_trash_folder):
            raise FileExistsError(f"{stack_trash_folder}")

        try:
            # 1. Download stack as multipage tif
            local_tif = self.download_files_as_single_tif(
                tif_files=stack_tifs,
                file_name=f"{stack_name}.tif",
                local_folder=local_temp_folder,
            )

            # 2. Upload multipage tif
            self.logger.info(f"Uploading {local_tif}")
            self.upload_file(local_tif, remote_tif)
            self.logger.info(f"Uploaded {local_tif} as {remote_tif}")

            # 3. Delete local multipage tif
            os.remove(local_tif)
            self.logger.info(f"Removed {local_tif}")

            # 4. Move stack tifs to trash
            self.create_empty_folder(stack_trash_folder)
            self.logger.info(f"Moving stack tifs to {stack_trash_folder}")
            for file in tqdm(stack_tifs):
                trash_path = stack_trash_folder / file.name
                self.move_file(file, trash_path)
        except (TifFileMissingException, TiffFileError) as e:
            self.logger.error(f"{e}")
            self.logger.info("Moving tifs into separate folder instead of combining.")
            incomplete_target = remote_folder / f"INCOMPLETE_{stack_name}"
            self.create_empty_folder(incomplete_target)
            self.logger.info(f"Moving stack tifs to {incomplete_target}")
            for file in tqdm(stack_tifs):
                new_path = incomplete_target / file.name
                self.move_file(file, new_path)

    def check_folder_structure(self, remote_folder: Path) -> tuple[dict, list]:
        """
        Check if a folder contains .tif files belonging to more than 1 stack.
        Also return stacks and tifs because scanning large folders takes time.
        """
        tifs = self.find_tifs_in_folder(remote_folder)
        if len(tifs) == 0:
            self.logger.info(f"No tifs in {remote_folder}.")
            stacks = {}
            single_tifs = []
        elif len(tifs) == 1:
            self.logger.info(f"Only 1 tif: {tifs[0]}.")
            stacks = {}
            single_tifs = tifs
        else:
            self.logger.info(f"{len(tifs)} tifs in {remote_folder} - looking for stacks.")
            stacks, single_tifs = self.identify_stacks_in_tifs(tifs)
        return stacks, single_tifs

    def find_tifs_in_folder(self, remote_folder: Path) -> list[Path]:
        """Find tif files in a nas folder."""
        files, folders = self.list_files_and_folders(remote_folder)
        tifs = self.identify_tifs_in_files(files)
        return tifs

    @staticmethod
    def identify_tifs_in_files(files: list[Path]) -> list[Path]:
        """Identify tif files in a list of files."""
        tifs = []
        for file in files:
            if file.suffix in [".tif", ".tiff"]:
                tifs.append(file)
        return tifs

    def identify_stacks_in_tifs(self, tifs: list[Path]) -> tuple[dict, list]:
        """Identify if tifs come from multi-frame recordings or if they are single snapshots.."""
        tif_origin = identify_tif_origin(tifs[0])
        if tif_origin == "leica":
            self.logger.info(f"Tifs (e.g. {tifs[0].name}) identified as leica tifs")
            stacks, single_tifs = self.identify_leica_stacks_in_tifs(tifs)
        elif tif_origin == "basler":
            self.logger.info(f"Tifs (e.g. {tifs[0].name}) identified as basler tifs")
            stacks, single_tifs = self.identify_basler_stacks_in_tifs(tifs)
        elif tif_origin == "leica_metadata":
            self.logger.info(f"{tifs[0].parent} identified as leica metadata tifs.")
            single_tifs = tifs
            stacks = {}
        else:
            self.logger.info(f"Could not identify stacks: neither leica nor basler.")
            single_tifs = tifs
            stacks = {}
        self.logger.info(f"{len(stacks)} tif stacks and {len(single_tifs)} single tifs found.")
        return stacks, single_tifs

    def identify_leica_stacks_in_tifs(self, tifs: list[Path]) -> tuple[dict, list]:
        """Sort tif files into stacks belonging to a single recording."""
        stacks = {}
        single_tifs = []
        for file in tifs:
            file_name = file.name
            findings = re.findall("_[zt][0-9]+_", file_name)
            if findings:
                splitter = findings[0]
                parts = file_name.split(splitter)
                stack_name = parts[0]
                if stack_name not in stacks.keys():
                    self.logger.info(f"Stack in {tifs[0].parent}: {stack_name}")
                    stacks[stack_name] = []
                stacks[stack_name].append(file)
            else:
                single_tifs.append(file)
        keys_to_delete = []
        for stack_name, stack_list in stacks.items():
            if len(stack_list) == 1:
                single_tifs.append(stack_list[0])
                keys_to_delete.append(stack_name)
                self.logger.info(f"Only 1 tif in stack {stack_name} -> Appending to single tifs.")
        for key in keys_to_delete:
            del stacks[key]
        return stacks, single_tifs

    @staticmethod
    def identify_basler_stacks_in_tifs(tifs: list[Path]) -> tuple[dict, list]:
        """Sort tif files into stacks belonging to a single recording."""
        stacks = {}
        single_tifs = []
        for file in tifs:
            parts = file.name.split("_")
            index_part = parts[-1]
            stack_name = "_".join(parts[:-1])
            if stack_name not in stacks.keys():
                stacks[stack_name] = []
            stacks[stack_name].append(file)
        return stacks, single_tifs

    def create_folder_for_stack(self, stack_name: str, remote_folder: Path, stack_files: list[Path]) -> None:
        """Move .tif files belonging to a single stack into a dedicated folder."""
        self.logger.info(f"{len(stack_files)} tifs in stack {stack_name}.")
        parent_directory = remote_folder / stack_name
        if parent_directory.is_dir():
            raise FileExistsError(f"{parent_directory}")
        self.connection.mkdir(str(parent_directory))
        for file in tqdm(stack_files):
            new_path = parent_directory / file.name
            self.move_file(file, new_path)
        self.logger.info(f"Moved {len(stack_files)} tifs into {parent_directory}.")

    def download_files_as_single_tif(self, tif_files: list, file_name: str, local_folder: Path) -> Path:
        temp_local = local_folder / "temp.tif"
        multipage_local = local_folder / file_name
        assert multipage_local.suffix == ".tif"

        n_files = len(tif_files)
        self.logger.info(f"{len(tif_files)} tif files")
        assert n_files > 1, f"Not enough files: {n_files=} < 1"
        assert_single_stack(tif_files)
        tif_files = sort_tif_files(tif_files)
        for i_file in tqdm(range(n_files)):
            file = tif_files[i_file]
            self.download_file(file, temp_local, overwrite=True)
            image = tifffile.imread(temp_local)
            if i_file == 0:
                shape = (n_files, image.shape[0], image.shape[1])
                tifffile.imwrite(
                    multipage_local,
                    shape=shape,
                    dtype=np.uint8,
                )
                store = tifffile.imread(multipage_local, mode="r+", aszarr=True)
                z = zarr.open(store, mode="r+")
                self.logger.info(f"Empty tif created: {shape}")
            z[i_file, :, :] = image
        os.remove(local_folder / "temp.tif")
        return multipage_local

    def download_folder_as_single_tif(self, remote_folder: Path, local_folder: Path) -> Path:
        """Download each tif of a stack, save into a singe local tif."""
        temp_local = local_folder / "temp.tif"
        multipage_local = local_folder / f"{remote_folder.name}.tif"
        tif_files = self.find_tifs_in_folder(remote_folder)
        n_files = len(tif_files)
        self.logger.info(f"{len(tif_files)} tif files in {remote_folder}")
        assert n_files > 1, f"Not enough files in {remote_folder}: {n_files} < 1"
        assert_single_stack(tif_files)
        tif_files = sort_tif_files(tif_files)
        for i_file in tqdm(range(n_files)):
            file = tif_files[i_file]
            self.download_file(file, temp_local, overwrite=True)
            image = tifffile.imread(temp_local)
            if i_file == 0:
                shape = (n_files, image.shape[0], image.shape[1])
                tifffile.imwrite(
                    multipage_local,
                    shape=shape,
                    dtype=np.uint8,
                )
                store = tifffile.imread(multipage_local, mode="r+", aszarr=True)
                z = zarr.open(store, mode="r+")
                self.logger.info(f"Empty tif created: {shape}")
            z[i_file, :, :] = image
        os.remove(local_folder / "temp.tif")
        return multipage_local



"""Basic connection to a network-associated server (NAS)."""
import logging
import os
import sys
from pathlib import Path

import pysftp
from tqdm import tqdm


class SftpNas:
    """Basic SFTP connection (wrapper around pysftp)"""

    def __init__(self, host_name: str, user_name: str, pwd: str, log_file: Path | None = None) -> None:
        """Create pysftp.Connection to server."""
        self._host_name = host_name
        self._user_name = user_name
        self._pwd = pwd
        self.log_file = log_file
        self.logger = self.create_logger()
        self.connection: pysftp.Connection = pysftp.Connection(
            host=host_name,
            username=user_name,
            password=pwd,
        )
        self.logger.info(f"Connection established: {host_name}@{user_name}")

    def create_logger(self) -> logging.Logger:
        logger = logging.Logger("nas")
        logger.setLevel(logging.DEBUG)
        stream_handler = logging.StreamHandler(stream=sys.stdout)
        formatter = logging.Formatter("%(asctime)s - %(message)s")
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
        if isinstance(self.log_file, Path):
            file_handler = logging.FileHandler(self.log_file)
            formatter = logging.Formatter("%(asctime)s - %(message)s")
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
            logger.info(f"Logging into {self.log_file}")
        return logger

    def close(self) -> None:
        """Close connection."""
        self.connection.close()

    def reconnect(self) -> None:
        try:
            self.connection.close()
        except Exception as e:
            self.logger.info(f"Could not close connection: {e}")
        self.connection: pysftp.Connection = pysftp.Connection(
            host=self._host_name,
            username=self._user_name,
            password=self._pwd,
        )

    def list_contents(self, remote_folder: Path) -> list[Path]:
        """List contents of a nas folder"""
        if not self.is_dir(remote_folder):
            raise FileNotFoundError(f"{remote_folder} not found.")
        contents: list[str] = self.connection.listdir(str(remote_folder))
        contents: list[Path] = [remote_folder / name for name in contents]
        return contents

    def is_dir(self, remote_path: Path) -> bool:
        """Check whether a nas path is a directory."""
        return self.connection.isdir(str(remote_path))

    def list_files_and_folders(self, remote_folder: Path) -> tuple:
        """List contents of a nas folder sorted into files and folders"""
        files = []
        folders = []
        contents = self.list_contents(remote_folder)
        for element in contents:
            if self.is_file(element):
                files.append(element)
            elif self.is_dir(element):
                folders.append(element)
            else:
                raise ValueError(f"{element} is neither file nor folder?")
        return files, folders

    def is_file(self, remote_path: Path) -> bool:
        """Check whether a nas path is a file."""
        return self.connection.isfile(str(remote_path))

    def download_file(self, remote_file: Path, local_file: Path, overwrite: bool = False) -> None:
        """Download a single file from the server"""
        if not self.is_file(remote_file):
            raise FileNotFoundError(f"{remote_file} does not exist.")
        if not local_file.parent.is_dir():
            raise FileNotFoundError(f"Target parent folder {local_file.parent} does not exist.")
        if local_file.is_file() and not overwrite:
            raise FileExistsError(f"{local_file} already exists.")
        self.connection.get(str(remote_file), str(local_file))

    def upload_file(self, local_file: Path, remote_file: Path) -> None:
        """Upload a single file to the server"""
        if not local_file.is_file():
            raise FileNotFoundError(f"{local_file} does not exist.")
        if not self.is_dir(remote_file.parent):
            raise FileNotFoundError(f"Target parent folder {remote_file.parent} does not exist.")
        if self.is_file(remote_file):
            raise FileExistsError(f"{remote_file} already exists.")
        self.connection.put(str(local_file), str(remote_file))

    def download_folder(self, remote_folder: Path, local_parent: Path, recursive: bool = True) -> Path:
        """Download a whole folder from the server"""
        if not self.is_dir(remote_folder):
            raise FileNotFoundError(f"{remote_folder} does not exist.")
        if not local_parent.is_dir():
            raise FileNotFoundError(f"Target parent folder {local_parent} does not exist.")
        local_folder = local_parent / remote_folder.name
        if local_folder.is_dir():
            raise FileExistsError(f"{local_folder} already exists.")
        os.mkdir(local_folder)
        files, folders = self.list_files_and_folders(remote_folder)
        self.logger.info(f"{len(files)} files found in {remote_folder}.")
        self.logger.info(f"{len(folders)} folders found in {remote_folder}.")
        for remote_file in tqdm(files):
            local_file = local_folder / remote_file.name
            self.download_file(remote_file, local_file)
        if recursive:
            for remote_sub_folder in folders:
                self.download_folder(remote_sub_folder, local_folder, recursive=recursive)
        return local_folder

    def upload_folder(self, local_source: Path, remote_parent: Path, recursive: bool = True, make_tree: bool = False) -> Path:
        """Upload a folder to the NAS"""
        if not local_source.is_dir():
            raise FileNotFoundError(f"{local_source} does not exist.")
        if not self.is_dir(remote_parent):
            if make_tree:
                self.connection.makedirs(str(remote_parent))
                self.logger.info(f"Folder tree to {remote_parent} created.")
            else:
                raise FileNotFoundError(f"{remote_parent} does not exist.")

        # create nas folder
        remote_target = remote_parent / local_source.name
        if self.connection.isdir(str(remote_target)):
            raise FileExistsError(f"{remote_target} already exists.")
        self.connection.mkdir(str(remote_target))
        self.logger.info(f"Remote folder {remote_target} created.")

        # get contents of local folder
        files, folders = self.list_local_files_and_folders(local_source)

        # upload files
        for local_file in files:
            remote_file = remote_target / local_file.name
            self.upload_file(local_file, remote_file)
            self.logger.info(f"{local_file} -> {remote_file}")

        # upload sub-folders
        if recursive:
            for local_sub_folder in folders:
                self.upload_folder(local_sub_folder, remote_target)
        return remote_target

    def print_tree(self, remote_folder: Path, recursive: bool = True, level: int = 0, max_level: int = 3) -> None:
        """Print file tree."""
        if level < max_level:
            files, folders = self.list_files_and_folders(remote_folder)
            indent = f"\t" * level
            for file in files:
                self.logger.info(f"{indent} {file.name}")

            for folder in folders:
                self.logger.info(f"{indent} {folder.name}")
                if recursive:
                    self.print_tree(folder, level=level + 1, max_level=max_level)

    def list_local_files_and_folders(self, local_folder: Path) -> tuple:
        """List local files and folders."""
        elements = local_folder.iterdir()
        files = []
        folders = []
        for element in elements:
            if element.is_file():
                files.append(element)
            elif element.is_dir():
                folders.append(element)
            else:
                raise ValueError(f"{element} is neither file nor folder?")
        return files, folders

    def move_file(self, remote_source: Path, remote_target: Path) -> None:
        """Move a file around."""
        if not self.is_file(remote_source):
            raise FileNotFoundError(f"{remote_source}")
        if self.is_file(remote_target):
            raise FileExistsError(f"{remote_target}")
        if not self.is_dir(remote_target.parent):
            raise FileNotFoundError(f"{remote_target.parent}")
        self.connection.rename(str(remote_source), str(remote_target))

    def move_folder(self, remote_source: Path, remote_target: Path) -> None:
        """Move a folder from one nas location to another."""
        if not self.is_dir(remote_source):
            raise FileNotFoundError(f"{remote_source} does not exist.")
        if self.is_dir(remote_target):
            raise FileExistsError(f"{remote_target} already exists.")
        self.connection.rename(str(remote_source), str(remote_target))

    def create_empty_folder(self, remote_folder: Path) -> None:
        """Create empty folder on NAS."""
        if self.is_dir(remote_folder):
            raise FileExistsError(f"{remote_folder} already exists.")
        self.connection.makedirs(str(remote_folder))

    def get_file_size(self, remote_file: Path, unit: str = "gb") -> float:
        """Get size of a remote file."""
        stats = self.connection.stat(str(remote_file))
        file_size = stats.st_size
        if unit == "gb":
            file_size = file_size / 10 ** 9
        else:
            raise NotImplementedError(f"{unit=} not implemented")
        return file_size






import socket
from pathlib import Path


def get_path_to_base() -> Path:
    host_name = socket.gethostname()
    if host_name == "mathis-Precision-Tower-3431":
        path = Path("/home/mathis/Code/gitlab/labnas")
    elif host_name == "matt-bluechip-BUSINESSline-individu":
        path = Path("/home/matt/Code/labnas")
    else:
        raise ValueError(f"{host_name=} unknown")
    return path
